# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for optimizers."""
from absl import logging
from absl.testing import parameterized
import jax
from jax import numpy as jnp
import optax
from tensorflow_probability.substrates import jax as tfp
import tree
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax.optimizers import optimizers

from absl.testing import absltest

tfb = tfp.bijectors


def loss_fn(inputs):
  # Sinusoidal function minimized at all inputs being zero. (min value=-2)
  # There are lots of local optima between [-3, 3]
  xs = jnp.concatenate(jax.tree_util.tree_flatten(inputs)[0], axis=-1)
  xs = jax.nn.sigmoid(xs)
  return jnp.mean(jnp.cos(xs * 5 * 2 * jnp.pi) * (2 - 5 *
                                                  (xs - 0.5)**2)), dict()


def optimizer_setup(rng):
  rng1, rng2 = jax.random.split(rng, 2)
  return {
      'x1': jax.random.uniform(rng1, shape=(1,), minval=-3, maxval=3),
      'x2': jax.random.uniform(rng2, shape=(2,), minval=-3., maxval=3.)
  }


def _make_constraint_array(b):
  if b is None:
    return None
  return {'x1': jnp.array([b]), 'x2': jnp.array([b, b])}


class OptimizersTest(parameterized.TestCase):

  @parameterized.parameters(
      (None,),
      ((-4.0, None),),
      ((None, 5.0),),
      ((-3.0, 3.0),),
  )
  def test_sinusoidal(self, bounds):
    optimize = optimizers.OptaxTrainWithRandomRestarts(
        optax.adam(5e-2), epochs=100, verbose=True, random_restarts=200)
    if bounds is None:
      constraints = None
    else:
      lb, ub = tree.map_structure(_make_constraint_array, bounds)
      constraints = sp.Constraint.create((lb, ub), tfb.SoftClip)

    optimal_params, metrics = optimize(
        optimizer_setup, loss_fn, jax.random.PRNGKey(1), constraints=constraints
    )
    logging.info('Optimal: %s', optimal_params)
    # Thanks to restarts, we can find the optimum.
    self.assertLessEqual(jnp.abs(loss_fn(optimal_params)[0] + 2), 5e-3)
    self.assertTrue((metrics['loss'][-1] < metrics['loss'][0]).all())
    if bounds is not None:
      for y_, b_ in zip(tree.flatten(optimal_params), tree.flatten(lb)):
        if b_ is not None:
          self.assertTrue((y_ > b_).all())
      for y_, b_ in zip(tree.flatten(optimal_params), tree.flatten(ub)):
        if b_ is not None:
          self.assertTrue((y_ < b_).all())

  @parameterized.parameters(
      (1, 1),
      (1, 5),
      (5, 20),
  )
  def test_sinusodial_bestn_optax(self, best_n, random_restarts):
    optimize = optimizers.OptaxTrainWithRandomRestarts(
        optax.adam(5e-3),
        epochs=100,
        verbose=True,
        random_restarts=random_restarts,
        best_n=best_n,
    )
    constraint_fn = tfb.JointMap({'x1': tfb.Exp(), 'x2': tfb.Softplus()})
    optimal_params, _ = optimize(
        optimizer_setup,
        loss_fn,
        jax.random.PRNGKey(0),
        # Optax uses the bijector and not the bounds, so it is safe to pass only
        # the bijector.
        constraints=sp.Constraint(bijector=constraint_fn),
    )
    logging.info('Optimal: %s', optimal_params)

    batch_shape = (best_n,) if best_n > 1 else ()
    self.assertSequenceEqual(optimal_params['x2'].shape, batch_shape + (2,))
    self.assertSequenceEqual(optimal_params['x1'].shape, batch_shape + (1,))

  @parameterized.parameters(
      (None,),
      ((-4.0, None),),
      ((None, 5.0), False),
      ((-3.0, 3.0),),
  )
  def test_sinusodial_bestn_scipy_l_bfgs_b(self, bounds, nest_constraint=True):
    if bounds is None:
      constraints = None
    else:
      if nest_constraint:
        bounds = tree.map_structure(_make_constraint_array, bounds)
      constraints = sp.Constraint.create(bounds, tfb.SoftClip)
    optimize = optimizers.JaxoptLbfgsB(
        random_restarts=6, best_n=5, use_scipy=True
    )
    optimal_params, _ = optimize(
        optimizer_setup, loss_fn, jax.random.PRNGKey(0), constraints=constraints
    )
    logging.info('Optimal: %s', optimal_params)

    self.assertSequenceEqual(optimal_params['x2'].shape, (5, 2))
    self.assertSequenceEqual(optimal_params['x1'].shape, (5, 1))
    if bounds is not None:
      for y_, b_ in zip(tree.flatten(optimal_params), tree.flatten(bounds[0])):
        if b_ is not None:
          self.assertTrue((y_ > b_).all())
      for y_, b_ in zip(tree.flatten(optimal_params), tree.flatten(bounds[1])):
        if b_ is not None:
          self.assertTrue((y_ < b_).all())

  @parameterized.parameters(
      (None,),
      ((-4.0, None),),
      ((None, 5.0), False),
      ((-3.0, 3.0),),
  )
  def test_sinusodial_bestn_jax_l_bfgs_b(self, bounds, nest_constraint=True):
    if bounds is None:
      constraints = None
    else:
      if nest_constraint:
        bounds = tree.map_structure(_make_constraint_array, bounds)
      constraints = sp.Constraint.create(bounds, tfb.SoftClip)
    optimize = optimizers.JaxoptLbfgsB(
        random_restarts=6, best_n=5, use_scipy=False
    )
    optimal_params, _ = optimize(
        optimizer_setup, loss_fn, jax.random.PRNGKey(0), constraints=constraints
    )
    logging.info('Optimal: %s', optimal_params)

    self.assertSequenceEqual(optimal_params['x2'].shape, (5, 2))
    self.assertSequenceEqual(optimal_params['x1'].shape, (5, 1))
    if bounds is not None:
      for y_, b_ in zip(tree.flatten(optimal_params), tree.flatten(bounds[0])):
        if b_ is not None:
          self.assertTrue((y_ > b_).all())
      for y_, b_ in zip(tree.flatten(optimal_params), tree.flatten(bounds[1])):
        if b_ is not None:
          self.assertTrue((y_ < b_).all())


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Benchmarking utilities."""

from vizier._src.benchmarks.analyzers.convergence_curve import ConvergenceComparatorBase
from vizier._src.benchmarks.analyzers.convergence_curve import ConvergenceCurve
from vizier._src.benchmarks.analyzers.convergence_curve import ConvergenceCurveConverter
from vizier._src.benchmarks.analyzers.convergence_curve import HypervolumeCurveConverter
from vizier._src.benchmarks.analyzers.convergence_curve import LogEfficiencyConvergenceCurveComparator
from vizier._src.benchmarks.analyzers.convergence_curve import PercentageBetterConvergenceCurveComparator
from vizier._src.benchmarks.analyzers.state_analyzer import BenchmarkStateAnalyzer
from vizier._src.benchmarks.runners.benchmark_runner import BenchmarkRunner
from vizier._src.benchmarks.runners.benchmark_runner import BenchmarkSubroutine
from vizier._src.benchmarks.runners.benchmark_runner import EvaluateActiveTrials
from vizier._src.benchmarks.runners.benchmark_runner import FillActiveTrials
from vizier._src.benchmarks.runners.benchmark_runner import GenerateAndEvaluate
from vizier._src.benchmarks.runners.benchmark_runner import GenerateSuggestions
from vizier._src.benchmarks.runners.benchmark_state import BenchmarkState
from vizier._src.benchmarks.runners.benchmark_state import BenchmarkStateFactory
from vizier._src.benchmarks.runners.benchmark_state import DesignerBenchmarkStateFactory
from vizier._src.benchmarks.runners.benchmark_state import ExperimenterDesignerBenchmarkStateFactory
from vizier._src.benchmarks.runners.benchmark_state import PolicyBenchmarkStateFactory
from vizier._src.benchmarks.runners.benchmark_state import PolicySuggester
from vizier._src.benchmarks.runners.benchmark_state import SeededPolicyFactory



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Analyzers for BenchmarkStates for fast comparisons and statistics."""

from typing import List

from vizier._src.benchmarks.analyzers import convergence_curve
from vizier._src.benchmarks.runners import benchmark_state


class BenchmarkStateAnalyzer:
  """Analyzer for BenchmarkStates."""

  @classmethod
  def to_curve(
      cls,
      states: List[benchmark_state.BenchmarkState],
      flip_signs_for_min: bool = False,
  ) -> convergence_curve.ConvergenceCurve:
    """Generates a ConvergenceCurve from a batch of BenchmarkStates.

    Each state in batch should represent the same study (different repeat).

    Args:
      states: List of BenchmarkStates.
      flip_signs_for_min: If true, flip signs of curve when it is MINIMIZE
        metric.

    Returns:
      Convergence curve with batch size equal to length of states.

    Raises:
      ValueError: When problem statements are not the same or is multiobjective.
    """
    if not states:
      raise ValueError('Empty States.')

    problem_statement = states[0].experimenter.problem_statement()
    if not problem_statement.is_single_objective:
      raise ValueError('Multiobjective Conversion not supported yet.')

    converter = convergence_curve.ConvergenceCurveConverter(
        problem_statement.metric_information.item(),
        flip_signs_for_min=flip_signs_for_min,
    )
    curves = []
    for state in states:
      if problem_statement != state.experimenter.problem_statement():
        raise ValueError(
            f'States must have same problem {problem_statement}'
            f' and {state.experimenter.problem_statement()}'
        )

      state_trials = state.algorithm.supporter.GetTrials()
      curve = converter.convert(state_trials)
      curves.append(curve)
    return convergence_curve.ConvergenceCurve.align_xs(curves)[0]



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Random vectorized optimizer to be used in convergence test."""

from typing import Optional

import jax
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import types
from vizier.pyvizier import converters


class RandomVectorizedStrategy(vb.VectorizedStrategy):
  """Random vectorized strategy."""

  def __init__(
      self,
      converter: converters.TrialToArrayConverter,
      suggestion_batch_size: int,
  ):
    self._converter = converter
    self._suggestion_batch_size = suggestion_batch_size
    self._n_features = sum(
        [spec.num_dimensions for spec in self._converter.output_specs]
    )

  def init_state(
      self,
      seed: jax.random.KeyArray,
      prior_features: Optional[types.Array] = None,
      prior_rewards: Optional[types.Array] = None,
  ) -> None:
    del seed
    return

  def suggest(self, state: None, seed: jax.random.KeyArray) -> jax.Array:
    del state
    return jax.random.uniform(
        seed,
        shape=(
            self._suggestion_batch_size,
            self._n_features,
        ),
    )

  def suggestion_batch_size(self) -> int:
    return self._suggestion_batch_size

  def update(
      self,
      state: None,
      batch_features: types.Array,
      batch_rewards: types.Array,
      seed: jax.random.KeyArray,
  ) -> None:
    return


def random_strategy_factory(
    converter: converters.TrialToArrayConverter,
    suggestion_batch_size: int,
) -> vb.VectorizedStrategy:
  """Creates a new vectorized strategy based on the Protocol."""
  return RandomVectorizedStrategy(
      converter=converter,
      suggestion_batch_size=suggestion_batch_size,
  )


def create_random_optimizer(
    converter: converters.TrialToArrayConverter,
    max_evaluations: int,
    suggestion_batch_size: int,
) -> vb.VectorizedOptimizer:
  """Creates a random optimizer."""
  return vb.VectorizedOptimizerFactory(
      strategy_factory=random_strategy_factory,
      max_evaluations=max_evaluations,
      suggestion_batch_size=suggestion_batch_size,
  )(converter=converter)


def create_random_optimizer_factory(
    max_evaluations: int, suggestion_batch_size: int
) -> vb.VectorizedOptimizerFactory:
  """Creates a random optimizer factory."""
  return vb.VectorizedOptimizerFactory(
      strategy_factory=random_strategy_factory,
      max_evaluations=max_evaluations,
      suggestion_batch_size=suggestion_batch_size,
  )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utils for vectorized eagle strategy."""

from typing import Optional

from flax import struct
import jax
from jax import numpy as jnp
from vizier.pyvizier import converters


@struct.dataclass
class EagleParamHandler:
  """Vectorized eagle strategy utils.

  The class is used to account for the different types of Vizier parameters and
  incorporate their naunces into the vectorized eagle strategy.

  Attributes:
    n_feature_dimensions: The total number of feature indices.
    n_categorical: The number of CATEGORICAL associated indices.
    has_categorical: A flag indicating if at least one feature is categorical.
    perturbation_factors: Array of continuous and categorical perturbation
      factors.
    _categorical_param_mask: A 2D array (n_categorical, n_feature_dimensions)
      that for each categorical parameters (row) has 1s in its associated
      feature indices. The array is used for sampling categorical values.
    _categorical_mask: A 1D array (n_feature_dimensions,) with 1s in the indices
      of categorical features and otherwise 0s. The array is used for sampling
      categorical values.
    _tiebreak_mask: A 1D array (n_feature_dimensions,) with multiplies of
      epsilons used to tie breaking. The array is used for sampling categorical
      values.
    _oov_mask: A 1D array (n_feature_dimensions,) with 1s in the non-oov
      indices. The array is used to generate random features with 0 value in the
      OOV indices.
    _epsilon: A small value used in tie-breaker
  """
  # Internal variables
  perturbation_factors: jax.Array
  _categorical_params_mask: jax.Array
  _categorical_mask: jax.Array
  _tiebreak_array: jax.Array
  _oov_mask: Optional[jax.Array]
  _epsilon: float
  # Public variables created by the class
  n_feature_dimensions: int = struct.field(pytree_node=False)
  n_categorical: int = struct.field(pytree_node=False)
  has_categorical: bool = struct.field(pytree_node=False)

  @classmethod
  def build(
      cls,
      converter: converters.TrialToArrayConverter,
      categorical_perturbation_factor: float,
      pure_categorical_perturbation_factor: float,
      epsilon: float = 1e-5,
  ) -> 'EagleParamHandler':
    """Docstring."""

    valid_types = [
        converters.NumpyArraySpecType.ONEHOT_EMBEDDING,
        converters.NumpyArraySpecType.CONTINUOUS
    ]
    unsupported_params = sum(
        1 for spec in converter.output_specs if spec.type not in valid_types
    )
    if unsupported_params:
      raise ValueError('Only CATEGORICAL/CONTINUOUS parameters are supported!')

    n_feature_dimensions = converter.to_features([]).shape[-1]
    n_categorical = sum(
        1
        for spec in converter.output_specs
        if spec.type == converters.NumpyArraySpecType.ONEHOT_EMBEDDING
    )
    has_categorical = n_categorical > 0
    all_features_categorical = n_feature_dimensions == n_categorical
    oov_mask = None
    tiebreak_array = None
    categorical_mask = None
    categorical_params_mask = None
    if has_categorical:
      categorical_params_mask = jnp.zeros((n_categorical, n_feature_dimensions))
      oov_mask = jnp.ones((n_feature_dimensions,))
      row = 0
      col = 0
      # Create a flag to indicate if the converter uses OOV padding. If none of
      # the CATEGORICAL params use padding the 'oov_mask' is set to None.
      is_pad_oov = False
      for spec in converter.output_specs:
        if spec.type == converters.NumpyArraySpecType.ONEHOT_EMBEDDING:
          n_dim = spec.num_dimensions
          categorical_params_mask = categorical_params_mask.at[
              row, col : col + n_dim
          ].set(1.0)
          if spec.num_oovs:
            oov_mask = oov_mask.at[col + n_dim - 1].set(0.0)
            is_pad_oov = True
          row += 1
          col += n_dim
        else:
          col += 1

      oov_mask = oov_mask if is_pad_oov else None
      tiebreak_array = -epsilon * jnp.arange(1, n_feature_dimensions + 1)
      categorical_mask = jnp.sum(categorical_params_mask, axis=0)

    perturbation_factors = []

    if all_features_categorical:
      for spec in converter.output_specs:
        perturbation_factors.extend(
            [pure_categorical_perturbation_factor] * spec.num_dimensions
        )
    else:
      for spec in converter.output_specs:
        if spec.type == converters.NumpyArraySpecType.ONEHOT_EMBEDDING:
          perturbation_factors.extend(
              [categorical_perturbation_factor] * spec.num_dimensions
          )

        elif spec.type == converters.NumpyArraySpecType.CONTINUOUS:
          perturbation_factors.append(1.0)
    # Add any extra dimensions at the end.
    perturbation_factors.extend(
        [0.0] * (n_feature_dimensions - len(perturbation_factors))
    )
    perturbation_factors = jnp.array(perturbation_factors)
    return EagleParamHandler(
        n_feature_dimensions=n_feature_dimensions,
        n_categorical=n_categorical,
        has_categorical=has_categorical,
        perturbation_factors=perturbation_factors,
        _categorical_params_mask=categorical_params_mask,
        _categorical_mask=categorical_mask,
        _tiebreak_array=tiebreak_array,
        _oov_mask=oov_mask,
        _epsilon=epsilon,
    )

  def sample_categorical(
      self, features: jax.Array, seed: jax.random.KeyArray
  ) -> jax.Array:
    """Sample categorical features.

    The categorical sampling is used before returning suggestion to ensure that
    only actual categorical values are suggested. Non categorical features are
    left unchanged.

    For example: If 'features' has one categorical parameter with 3 values and
    one float parameter the conversion will be take the form of:
    (0.1, 0.3, 0.5, 0.4578) -> (0, 1, 0, 0.4578).

    Implementation details:
    ----------------------
    1. For each categorical parameter, isolate its indices in a separate row.
    2. Normalize new row values to probabilities through dividing by the sum.
    3. Create cummulatitive sum of probabilites to generat a CDF.
    4. Add small incremental values to CDF to tie-break (explained more later).
    5. Randomize uniform values for each row to determine which value to sample.
    6. Find the minimum index that its CDF > uniform. See code around
    'sampled_categorical_params' below for more detail and for why we need to
    add the tie-breaking array values.
    7. Flatten each row sampled values and combine with original features.

    Arguments:
      features: (batch_size, n_feature_dimensions)
      seed: Random seed.

    Returns:
      The features with sampled categorical parameters.
        (batch_size, n_feature_dimensions)
    """
    if not self.has_categorical:
      return features
    batch_size = features.shape[0]
    # Mask each row (which represents a categorical param) to remove values in
    # indices that aren't associated with the parameter indices.
    param_features = features[:, jnp.newaxis, :] * self._categorical_params_mask
    # Create probabilities from non-normalized parameter features values.
    probs = param_features / jnp.sum(param_features, axis=-1, keepdims=True)
    # Generate random uniform values to use for sampling.
    # TODO: Pre-compute random values in batch to improve performance.
    unifs = jax.random.uniform(seed, shape=(batch_size, self.n_categorical, 1))
    # Find the locations of the indices that exceed random values.
    locs = jnp.cumsum(probs, axis=-1) >= unifs
    # Multiply by 'categorical_mask' to mask off cumsum in non-categorical
    # indices, and add 'tiebreak_mask' to find the first index.
    masked_locs = locs * self._categorical_params_mask + self._tiebreak_array
    # Generate the samples so that each parameter features has a single 1 value.
    sampled_categorical_params = jnp.trunc(
        masked_locs / jnp.max(masked_locs, axis=-1, keepdims=True)
    )
    # Flatten all the categories features to dimension
    # (batch_size, n_feature_dimensions)
    sampled_features = jnp.sum(sampled_categorical_params, axis=1)
    # Mask categorical features and add the new sampled categorical values.
    return sampled_features + features * (1 - self._categorical_mask)

  def random_features(
      self, batch_size: int, seed: jax.random.KeyArray
  ) -> jax.Array:
    """Create random features with uniform distribution.

    In case there are CATEGORICAL features with OOV we use the 'oov_mask' which
    is 1D numpy array (n_feature_dimensions,) with 0s in OOV indices and
    otherwise 1s.  After multiplying (and broadcasting) the randomly generated
    features with the mask we're guaranteed that no features will be created in
    OOV indices.  Therefore when mutating fireflies' features (index by index),
    the final suggested features will also have 0s in the OOV indices as
    desired.

    Arguments:
      batch_size:
      seed: Random seed.

    Returns:
      The random features with out of vocabulary indices zeroed out.
    """
    features = jax.random.uniform(
        seed, shape=(batch_size, self.n_feature_dimensions)
    )
    if self._oov_mask is not None:
      # Don't create random values for CATEGORICAL features OOV indices.
      # Broadcasting:
      #   (batch_size, n_feature_dimensions) x (n_feature_dimensions,)
      features = features * self._oov_mask
    return features



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Scalarizing Designer for MOO via reducing to a scalarized objective."""
import copy
from typing import Optional, Protocol, Sequence

import attr
import jax
from jax import numpy as jnp
from jax.typing import ArrayLike
from vizier import algorithms as vza
from vizier import pyvizier as vz


# TODO Separate into another file/folder.
class Scalarization(Protocol):
  """Reduces an array of objectives to a single float.

  Assumes all objectives are for MAXIMIZATION.
  """

  def __call__(self, objectives: ArrayLike) -> jax.Array:
    pass


@attr.define(init=True)
class HyperVolumeScalarization(Scalarization):
  """HyperVolume Scalarization."""

  weights: ArrayLike = attr.ib()

  reference_point: Optional[ArrayLike] = attr.ib(default=None, kw_only=True)

  def __attrs_post_init__(self):
    if any(self.weights <= 0):
      raise ValueError(f'Non-positive weights {self.weights}')

  def __call__(  # pytype: disable=signature-mismatch  # numpy-scalars
      self,
      objectives: ArrayLike,
  ) -> jax.Array:
    # Uses scalarizations in https://arxiv.org/abs/2006.04655 for
    # non-convex multiobjective optimization.
    if self.reference_point is not None:
      return jnp.min((objectives - self.reference_point) / self.weights)
    else:
      return jnp.min(objectives / self.weights)


class ScalarizingDesigner(vza.Designer):
  """Multiobjective Scalarizing Designer.

  This designer applies multiobjective optimization by reducing to
  single objective optimization via scalarization.
  """

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      designer_factory: vza.DesignerFactory[vza.Designer],
      scalarization: Scalarization,
      *,
      seed: Optional[int] = None,
  ):
    """Init.

    Args:
      problem_statement: Must be a mulitobjective search space.
      designer_factory: Factory to create the single-objective designer.
      scalarization: Scalarization to be applied to objective metrics.
      seed: Any valid seed for factory.
    """
    self._scalarization = scalarization
    self._objectives = problem_statement.metric_information.of_type(
        vz.MetricType.OBJECTIVE
    )
    if len(self._objectives) <= 1:
      raise ValueError(f'Problem should be multi-objective {self._objectives}')

    # Create a single-objective designer.
    self._scalarized_metric_name = 'scalarized'
    single_objective_metric = problem_statement.metric_information.exclude_type(
        vz.MetricType.OBJECTIVE
    )
    single_objective_metric.append(
        vz.MetricInformation(
            name=self._scalarized_metric_name,
            goal=vz.ObjectiveMetricGoal.MAXIMIZE,
        )
    )
    self._problem_statement = copy.deepcopy(problem_statement)
    self._problem_statement.metric_information = single_objective_metric
    self._designer = designer_factory(self._problem_statement, seed=seed)

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    for trial in completed.trials:
      objectives = [
          trial.final_measurement.metrics.get_value(
              config.name, default=jnp.nan
          )
          for config in self._objectives
      ]
      # Simply append the scalarized value.
      trial.final_measurement.metrics[self._scalarized_metric_name] = (
          self._scalarization(jnp.array(objectives))
      )

    self._designer.update(completed, all_active)

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    return self._designer.suggest(count)
